/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <gstfdsrc.h>


GstElementDetails gst_fdsrc_details = {
  "Filedescriptor Source",
  "Source",
  "Read from a file descriptor",
  VERSION,
  "Erik Walthinsen <omega@cse.ogi.edu>",
  "(C) 1999",
};


/* FdSrc signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_FD,
  ARG_LOCATION,
  ARG_BYTESPERREAD
};


static void gst_fdsrc_class_init(GstFdSrcClass *klass);
static void gst_fdsrc_init(GstFdSrc *fdsrc);
static void gst_fdsrc_set_arg(GtkObject *object,GtkArg *arg,guint id);
static void gst_fdsrc_get_arg(GtkObject *object,GtkArg *arg,guint id);


static GstSrcClass *parent_class = NULL;
static guint gst_fdsrc_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_fdsrc_get_type(void) {
  static GtkType fdsrc_type = 0;

  if (!fdsrc_type) {
    static const GtkTypeInfo fdsrc_info = {
      "GstFdSrc",
      sizeof(GstFdSrc),
      sizeof(GstFdSrcClass),
      (GtkClassInitFunc)gst_fdsrc_class_init,
      (GtkObjectInitFunc)gst_fdsrc_init,
      (GtkArgSetFunc)gst_fdsrc_set_arg,
      (GtkArgGetFunc)gst_fdsrc_get_arg,
      (GtkClassInitFunc)NULL,
    };
    fdsrc_type = gtk_type_unique(GST_TYPE_SRC,&fdsrc_info);
  }
  return fdsrc_type;
}

static void
gst_fdsrc_class_init(GstFdSrcClass *klass) {
  GtkObjectClass *gtkobject_class;
  GstSrcClass *gstsrc_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstsrc_class = (GstSrcClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_SRC);

  gtk_object_add_arg_type("GstFdSrc::fd", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_LOCATION);
  gtk_object_add_arg_type("GstFdSrc::location", GTK_TYPE_STRING,
                          GTK_ARG_WRITABLE, ARG_LOCATION);
  gtk_object_add_arg_type("GstFdSrc::bytesperread", GTK_TYPE_INT,
                          GTK_ARG_READWRITE, ARG_BYTESPERREAD);

  gtkobject_class->set_arg = gst_fdsrc_set_arg;
  gtkobject_class->get_arg = gst_fdsrc_get_arg;

  gstsrc_class->push = gst_fdsrc_push;
//  gstsrc_class->push_region = gst_fdsrc_push_region;
}

static void gst_fdsrc_init(GstFdSrc *fdsrc) {
  fdsrc->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(fdsrc),fdsrc->srcpad);

  fdsrc->fd = 0;
  fdsrc->size = 0;

  fdsrc->curoffset = 0;
  fdsrc->seq = 0;
  fdsrc->bytes_per_read = 4096;
}

GstElement *gst_fdsrc_new(gchar *name) {
  GstElement *fdsrc = GST_ELEMENT(gtk_type_new(GST_TYPE_FDSRC));
  gst_object_set_name(GST_OBJECT(fdsrc),name);
  return fdsrc;
}

GstElement *gst_fdsrc_new_with_fd(gchar *name,int fd) {
  GstElement *fdsrc = gst_fdsrc_new(name);
  gtk_object_set(GTK_OBJECT(fdsrc),"fd",fd,NULL);
  return fdsrc;
}

void gst_fdsrc_push(GstSrc *src) {
  GstFdSrc *fdsrc;
  GstBuffer *buf;
  glong readbytes;

  g_return_if_fail(src != NULL);
  g_return_if_fail(GST_IS_FDSRC(src));
  fdsrc = GST_FDSRC(src);

  buf = gst_buffer_new();
  g_return_if_fail(buf);
  buf->data = (gpointer)malloc(fdsrc->bytes_per_read);
  readbytes = read(fdsrc->fd,buf->data,fdsrc->bytes_per_read);
  if (readbytes == 0) {
    gst_src_signal_eof(GST_SRC(fdsrc));
    return;
  }

  buf->srcoffset = fdsrc->curoffset;
  buf->datasize = readbytes;
  fdsrc->curoffset += readbytes;

  gst_pad_push(fdsrc->srcpad,buf);
}

static void gst_fdsrc_set_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstFdSrc *fdsrc;
  int fd;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_FDSRC(object));
  fdsrc = GST_FDSRC(object);

  /* FIXME there should be some sanity checking here, maybe */
  switch(id) {
    case ARG_FD:
      fdsrc->fd = GTK_VALUE_INT(*arg);
      break;
    case ARG_LOCATION:
      if (sscanf(GTK_VALUE_STRING(*arg),"%d",&fd))
        fdsrc->fd = fd;
      break;
    case ARG_BYTESPERREAD:
      fdsrc->bytes_per_read = GTK_VALUE_INT(*arg);
      break;
    default:
      break;
  }
}

static void gst_fdsrc_get_arg(GtkObject *object,GtkArg *arg,guint id) {
  GstFdSrc *fdsrc;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_FDSRC(object));
  fdsrc = GST_FDSRC(object);

  switch (id) {
    case ARG_FD:
      GTK_VALUE_INT(*arg) = fdsrc->fd;
      break;
    case ARG_BYTESPERREAD:
      GTK_VALUE_INT(*arg) = fdsrc->bytes_per_read;
      break;
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}
