/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <gst/gstobject.h>


/* Object signals and args */
enum {
  PARENT_SET,
  LAST_SIGNAL
};

enum {
  ARG_0,
  /* FILL ME */
};


static void gst_object_class_init(GstObjectClass *klass);
static void gst_object_init(GstObject *object);


static GstObject *parent_class = NULL;
static guint gst_object_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_object_get_type(void) {
  static GtkType object_type = 0;

  if (!object_type) {
    static const GtkTypeInfo object_info = {
      "GstObject",
      sizeof(GstObject),
      sizeof(GstObjectClass),
      (GtkClassInitFunc)gst_object_class_init,
      (GtkObjectInitFunc)gst_object_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    object_type = gtk_type_unique(gtk_object_get_type(),&object_info);
  }
  return object_type;
}

static void
gst_object_class_init(GstObjectClass *klass) {
  GtkObjectClass *gtkobject_class;

  gtkobject_class = (GtkObjectClass*)klass;

  parent_class = gtk_type_class(gtk_object_get_type());

  gst_object_signals[PARENT_SET] =
    gtk_signal_new("parent_set",GTK_RUN_LAST,gtkobject_class->type,
                   GTK_SIGNAL_OFFSET(GstObjectClass,parent_set),
                   gtk_marshal_NONE__POINTER,GTK_TYPE_NONE,1,
                   GTK_TYPE_POINTER);
  gtk_object_class_add_signals(gtkobject_class,gst_object_signals,LAST_SIGNAL);
}

static void gst_object_init(GstObject *object) {
  object->name = NULL;
  object->flags = 0;
  object->parent = NULL;
  object->lock = g_mutex_new();
}

/**
 * gst_object_new:
 *
 * Create a new object with
 *
 * Returns: new object
 */
GstObject *gst_object_new() {
  return GST_OBJECT(gtk_type_new(gst_object_get_type()));
}

/**
 * gst_object_ref:
 * @object: GstObject to reference
 *
 * Increment the reference count of the given object.
 */
void gst_object_ref(GstObject *object) {
  g_return_if_fail(object != NULL);
  g_return_if_fail(GST_IS_OBJECT(object));

  gtk_object_ref(GTK_OBJECT(object));
}

/**
 * gst_object_unref:
 * @object: GstObject to unreference
 *
 * Decrement the reference count of the given object, freeing it if the
 * count hits zero.
 */
void gst_object_unref(GstObject *object) {
  g_return_if_fail(object != NULL);
  g_return_if_fail(GST_IS_OBJECT(object));

  gtk_object_unref(GTK_OBJECT(object));
}

/**
 * gst_object_destroy:
 * @object: GstObject to destroy
 *
 * Destroy the given object, freeing its memory.
 */
void gst_object_destroy(GstObject *object) {
  g_return_if_fail(object != NULL);
  g_return_if_fail(GST_IS_OBJECT(object));

  gtk_object_destroy(GTK_OBJECT(object));
}

/**
 * gst_object_set_name:
 * @object: GstObject to set name of
 * @name: new name of object
 *
 * Set the name of the object, getting rid of the old name if there was
 * one.
 */
void gst_object_set_name(GstObject *object,gchar *name) {
  g_return_if_fail(object != NULL);
  g_return_if_fail(GST_IS_OBJECT(object));
  g_return_if_fail(name != NULL);

  if (object->name != NULL)
    g_free(object->name);

  object->name = g_strdup(name);
}

/**
 * gst_object_get_name:
 * @object: GstObject to set name of
 *
 * Get the name of the object.
 *
 * Returns: name of the object
 */
gchar *gst_object_get_name(GstObject *object) {
  g_return_if_fail(object != NULL);
  g_return_if_fail(GST_IS_OBJECT(object));

  return object->name;
}

/**
 * gst_object_set_parent:
 * @object: GstObject to set parent of
 * @parent: new parent of object
 *
 * Set the parent of the object.  The object's reference count is
 * incremented.
 */
void gst_object_set_parent(GstObject *object,GstObject *parent) {
  g_return_if_fail(object != NULL);
  g_return_if_fail(GST_IS_OBJECT(object));
  g_return_if_fail(object->parent == NULL);
  g_return_if_fail(parent != NULL);
  g_return_if_fail(GST_IS_OBJECT(parent));
  g_return_if_fail(object != parent);

  gst_object_ref(object);
  object->parent = parent;

  gtk_signal_emit(GTK_OBJECT(object),gst_object_signals[PARENT_SET],parent);
}

/**
 * gst_object_get_parent:
 * @object: GstObject to get parent of
 *
 * Return the parent of the object.
 *
 * Returns: parent of the object
 */
GstObject *gst_object_get_parent(GstObject *object) {
  g_return_if_fail(object != NULL);
  g_return_if_fail(GST_IS_OBJECT(object));

  return object->parent;
}

/**
 * gst_object_unparent:
 * @object: GstObject to unparent
 *
 * Clear the parent of the object, removing the associated reference.
 */
void gst_object_unparent(GstObject *object) {
  g_return_if_fail(object != NULL);
  g_return_if_fail(GST_IS_OBJECT(object));
  g_return_if_fail(object->parent == NULL);

  object->parent = NULL;
  gst_object_unref(object);
}
